;;; $Id: kScheme.scm,v 1.4 1996/09/21 09:34:18 queinnec Exp $

;;; Un petit evaluateur avec continuations explicites.
;;; Christian Queinnec
;;; Universite Paris 6 & INRIA-Rocquencourt.

(define (atom? e)
  (not (pair? e)) )

(define (evaluate e r k)
  (if (atom? e)
      (cond ((symbol? e) (k (lookup e r)))
            (else (k e)) )
      (let ((special (assq (car e) *specials*)))
        (if (pair? special)
            ((cdr special) e r k)
            (evaluate-application (car e) (cdr e) r k) ) ) ) )

(define *specials* '())

(define-macro (define-special call . body)
  (let ((e (gensym)))
    `(begin
       (set! *specials*
             (cons (cons ',(caar call)
                         (lambda (,e . ,(cdr call))
                           (apply (lambda ,(cdar call)
                                    . ,body )
                                  (cdr ,e) ) ) )
                   *specials* ) )
       ',(caar call) ) ) )

(define-special ((quote expression) r k)
  (k expression) )

(define-special ((if condition consequent . alternant) r k)
  (evaluate condition
            r
            (lambda (bool)
              (if bool 
                  (evaluate consequent r k)
                  (if (pair? alternant) 
                      (evaluate (car alternant) r k)
                      (k nothing-particular) ) ) ) ) )

(define-special ((begin . forms) r k)
  (evaluate-begin forms r k) )

(define (evaluate-begin forms r k)
  (if (pair? forms)
      (evaluate (car forms)
                r
                (if (pair? (cdr forms))
                    (lambda (v)
                      (evaluate-begin (cdr forms) r k) )
                    k ) )
      (k nothing-particular) ) )

(define-special ((set! variable form) r k)
  (evaluate form 
            r
            (lambda (v)
              (k (update! variable r v)) ) ) )

(define-special ((lambda variables . body) r k)
  (k (make-function variables body r)) )

(define (evaluate-application function arguments r k)
  (evaluate function
            r
            (lambda (f)
              (evaluate-arguments arguments
                                  r
                                  (lambda (v*)
                                    (f k v*) ) ) ) ) )

(define (evaluate-arguments arguments r k)
  (if (pair? arguments)
      (evaluate (car arguments)
                r
                (lambda (v)
                  (evaluate-arguments (cdr arguments)
                                      r
                                      (lambda (v*)
                                        (k (cons v v*)) ) ) ) )
      (k '()) ) )
   
;;; Functions

(define (make-function variables body r)
  (lambda (k arguments)
    (evaluate-begin body (extend r variables arguments) k) ) )

;;; Environments

(define (lookup variable r)
  (let ((binding (assq variable r)))
    (if (pair? binding)
        (cdr binding)
        (wrong "No such binding" variable r) ) ) )

(define (update! variable r value)
  (let ((binding (assq variable r)))
    (if (pair? binding)
        (let ((old (cdr binding)))
          (set-cdr! binding value)
          (if (eq? old uninitialized) value old) )
        (wrong "No such binding" variable r value) ) ) )

(define (extend r variables values)
  (cond
   ((pair? variables)
    (if (pair? values)
        (cons (cons (car variables) (car values))
              (extend r (cdr variables) (cdr values)) )
        (wrong "Missing argument" variables values) ) )
   ((null? variables)
    (if (pair? values)
        (wrong "Too much argument" values)
        r ) )
   ((symbol? variables)
    (cons (cons variables values) r) )
   (else (wrong "Incorrect variable" variables)) ) )

;;; L'environnement initial

(define r.init '())

(define uninitialized "UnInitialized")

(define-macro (definitial name . value)
  `(begin (set! r.init
                (cons (cons ',name 
                            ,(if (pair? value) (car value)
                                 `uninitialized ) )
                      r.init ) )
          ',name ) )

(define-macro (defprimitive name primitive arity)
  `(definitial ,name
     (lambda (k args)
       (if (= (length args) ,arity)
           (k (apply ,primitive args))
           (wrong "Incorrect arity for" ',name args) ) ) ) )

(define-macro (defnaryprimitive name primitive arity)
  `(definitial ,name
     (lambda (k args)
       (if (>= (length args) ,arity)
           (k (apply ,primitive args))
           (wrong "Incorrect arity for" ',name args) ) ) ) )

(defprimitive cons cons 2)
(defprimitive car car 1)
(defprimitive cdr cdr 1)
(defprimitive pair? pair? 1)
(defprimitive symbol? symbol? 1)
(defprimitive eq? eq? 2)
(defprimitive null? null? 1)
(defprimitive set-car! set-car! 2)
(defprimitive set-cdr! set-cdr! 2)
(defnaryprimitive + + 1)
(defprimitive - - 2 )
(defprimitive = = 2)
(defprimitive < < 2)
(defprimitive > > 2)
(defnaryprimitive * * 1)
(defprimitive <= <= 2)
(defprimitive  >= >= 2)
(defprimitive remainder remainder 2)

(definitial call/cc 
  (lambda (k args)
    (if (= (length args) 1)
        (let ((f (car args))
              (kf (lambda (kk v*)
                    (if (= (length v*) 1)
                        (k (car v*))
                        (wrong "Incorrect arity for a continuation") ) )) )
          (f k (list kf)) )
        (wrong "Incorrect arity for" 'call/cc) ) ) )

(definitial apply 
  (lambda (k values)
    (define (listify arguments)
      (if (pair? (cdr arguments))
          (cons (car arguments) (listify (cdr arguments)))
          (car arguments) ) )
    (if (>= (length values) 2)
        ((car values) k (listify (cdr values)))
        (wrong "Incorrect arity" 'apply) ) ) )

;;; Some fundamental constants:

(definitial t #t)
(definitial f #f)
(definitial nil '())

(defnaryprimitive list
   (lambda values values)
   0 ) 

;;; Some free global locations:

(definitial x 'void)
(definitial y 'void)
(definitial z 'void)
(definitial a 'void)
(definitial b 'void)
(definitial c 'void)
(definitial foo 'void)
(definitial bar 'void)
(definitial hux 'void)
(definitial fib 'void)
(definitial fact 'void)
(definitial visit 'void)
(definitial length 'void)
(definitial primes 'void)

;;; Divers

(define nothing-particular #t)

;;; Starting the interpreter

(define (kscheme)
  (interpreter "kScheme? " "kScheme= " #t
    (lambda (read print error)
      (set! wrong error)
      (lambda ()
         (evaluate (read) r.init print) ) ) ) )

(define (test-kscheme file)
  (suite-test file "kScheme? " "kScheme= " #t
   (lambda (read check error)
     (set! wrong error)
     (lambda () 
       (evaluate (read) 
                 r.init 
                 (lambda (v) 
                   (check v) ) ) ) )
   equal? ) )
        
;;; end of kScheme.scm
